#include <ctype.h>
#include "mocktcp.h"

#define X_MOCKTCP_PORT 10000

#if defined(SO_DONTLINGER) && defined(SO_LINGER)
#undef SO_DONTLINGER
#endif

#define MIN_BACKLOG 128
#ifdef SOMAXCONN
#if SOMAXCONN > MIN_BACKLOG
#define BACKLOG SOMAXCONN
#endif
#endif
#ifndef BACKLOG
#define BACKLOG MIN_BACKLOG
#endif
/*
 * This is the MockTCP implementation of the X Transport service layer.
 *
 */

/*
 * This function gets the local address of the socket and stores it in the
 * XtransConnInfo structure for the connection.
 */

static int
TRANS(MockTCPGetAddr) (ciptr)

XtransConnInfo ciptr;

{
    struct sockaddr_in 	sockname;
#if defined(SVR4) || defined(SCO325)
    size_t namelen = sizeof sockname;
#else
    int namelen = sizeof sockname;
#endif

    PRMSG (3,"MockTCPGetAddr(%x)\n", ciptr, 0, 0);

    if (getsockname (ciptr->fd,(struct sockaddr *) &sockname, &namelen) < 0)
    {
	PRMSG (1,"MockTCPGetAddr: getsockname() failed: %d\n",
	    EGET(),0, 0);
	return -1;
    }

    /*
     * Everything looks good: fill in the XtransConnInfo structure.
     */

    if ((ciptr->addr = (char *) xalloc (namelen)) == NULL)
    {
        PRMSG (1,
	    "MockTCPGetAddr: Can't allocate space for the addr\n",
	    0, 0, 0);
        return -1;
    }

    ciptr->family = sockname.sin_family;
    ciptr->addrlen = namelen;
    memcpy (ciptr->addr, &sockname, ciptr->addrlen);

    return 0;
}


/*
 * This function gets the remote address of the socket and stores it in the
 * XtransConnInfo structure for the connection.
 */

static int
TRANS(MockTCPGetPeerAddr) (ciptr)

XtransConnInfo ciptr;

{
    struct sockaddr_in 	sockname;
#if defined(SVR4) || defined(SCO325)
    size_t namelen = sizeof sockname;
#else
    int namelen = sizeof sockname;
#endif

    PRMSG (3,"MockTCPGetPeerAddr(%x)\n", ciptr, 0, 0);
	
    if (mt_getpeername (ciptr->fd, (struct sockaddr_in *) &sockname) < 0)
    {
	PRMSG (1,"MockTCPGetPeerAddr: getpeername() failed: %d\n",
	    EGET(), 0, 0);
	return -1;
    }
	

    /*
     * Everything looks good: fill in the XtransConnInfo structure.
     */

    if ((ciptr->peeraddr = (char *) xalloc (namelen)) == NULL)
    {
        PRMSG (1,
	   "MockTCPGetPeerAddr: Can't allocate space for the addr\n",
	   0, 0, 0);
        return -1;
    }

    ciptr->peeraddrlen = namelen;
    memcpy (ciptr->peeraddr, &sockname, ciptr->peeraddrlen);

    return 0;
}


static XtransConnInfo
TRANS(MockTCPOpen) (void)
{
    XtransConnInfo	ciptr;

    PRMSG (3,"MockTCPOpen()\n", 0, 0, 0);

    if ((ciptr = (XtransConnInfo) xcalloc (
	1, sizeof(struct _XtransConnInfo))) == NULL)
    {
	PRMSG (1, "MockTCPOpen: malloc failed\n", 0, 0, 0);
	return NULL;
    }

    if ((ciptr->fd = mt_socket())<0)
    {
	PRMSG (1, "MockTCPOpen: mt_socket() failed\n", 0, 0, 0);
	xfree ((char *) ciptr);
	return NULL;
    }

    return ciptr;
}


#ifdef TRANS_REOPEN

static XtransConnInfo
TRANS(MockTCPReopen) (int fd, char *port)
{
    XtransConnInfo	ciptr;

    PRMSG (3,"MockTCPReopen(%d,%s)\n", fd, port,0);

    if ((ciptr = (XtransConnInfo) xcalloc (
	1, sizeof(struct _XtransConnInfo))) == NULL)
    {
	PRMSG (1, "MockTCPReopen: malloc failed\n", 0, 0, 0);
	return NULL;
    }

    ciptr->fd = fd;

    return ciptr;
}

#endif /* TRANS_REOPEN */


/*
 * These functions are the interface supplied in the Xtransport structure
 */

#ifdef TRANS_CLIENT

static XtransConnInfo
TRANS(MockTCPOpenCOTSClient) (thistrans, protocol, host, port)

Xtransport *thistrans;
char 	   *protocol;
char 	   *host;
char       *port;

{
    XtransConnInfo	ciptr;
    int			i;

    PRMSG (2, "MockTCPOpenCOTSClient(%s,%s,%s)\n",
	protocol, host, port);

    if ((ciptr = TRANS(MockTCPOpen) ()) == NULL)
    {
	PRMSG (1,"MockTCPOpenCOTSClient: Unable to open socket", 0,
	    0, 0);
	return NULL;
    }

    /* Save the index for later use */

    ciptr->index = i;

    return ciptr;
}

#endif /* TRANS_CLIENT */


#ifdef TRANS_SERVER

static XtransConnInfo
TRANS(MockTCPOpenCOTSServer) (thistrans, protocol, host, port)

Xtransport *thistrans;
char 	   *protocol;
char 	   *host;
char 	   *port;

{
    XtransConnInfo	ciptr;
    PRMSG (2,"MockTCPOpenCOTSServer(%s,%s,%s)\n", protocol, host, port);

    if ((ciptr = TRANS(MockTCPOpen) ()) == NULL)
    {
	PRMSG (1,"MockTCPOpenCOTSServer: Unable to open socket for %s\n",
	    thistrans->TransName, 0, 0);
	return NULL;
    }

    /*
     * Using this prevents the bind() check for an existing server listening
     * on the same port, but it is required for other reasons.
     */

    return ciptr;
}

#endif /* TRANS_SERVER */


#ifdef TRANS_CLIENT

static XtransConnInfo
TRANS(MockTCPOpenCLTSClient) (thistrans, protocol, host, port)

Xtransport *thistrans;
char 	   *protocol;
char 	   *host;
char 	   *port;

{
    XtransConnInfo	ciptr;
    int			i;

    PRMSG (2,"MockTCPOpenCLTSClient(%s,%s,%s)\n", protocol, host, port);

    if ((ciptr = TRANS(MockTCPOpen) ()) == NULL)
    {
	PRMSG (1,"MockTCPOpenCLTSClient: Unable to open socket for %s\n",
	      thistrans->TransName, 0, 0);
	return NULL;
    }

    /* Save the index for later use */

    return ciptr;
}

#endif /* TRANS_CLIENT */


#ifdef TRANS_SERVER

static XtransConnInfo
TRANS(MockTCPOpenCLTSServer) (thistrans, protocol, host, port)

Xtransport *thistrans;
char 	   *protocol;
char 	   *host;
char 	   *port;

{
    XtransConnInfo	ciptr;
    int	i;

    PRMSG (2,"MockTCPOpenCLTSServer(%s,%s,%s)\n", protocol, host, port);

    if ((ciptr = TRANS(MockTCPOpen) ()) == NULL)
    {
	PRMSG (1,"MockTCPOpenCLTSServer: Unable to open socket for %s\n",
	      thistrans->TransName, 0, 0);
	return NULL;
    }

    /* Save the index for later use */

    return ciptr;
}

#endif /* TRANS_SERVER */


#ifdef TRANS_REOPEN

static XtransConnInfo
TRANS(MockTCPReopenCOTSServer) (thistrans, fd, port)

Xtransport *thistrans;
int	   fd;
char	   *port;

{
    XtransConnInfo	ciptr;
    int			i;

    PRMSG (2,
	"MockTCPReopenCOTSServer(%d, %s)\n", fd, port, 0);

    if ((ciptr = TRANS(MockTCPReopen) (fd, port)) == NULL)
    {
	PRMSG (1,
	    "MockTCPReopenCOTSServer: Unable to reopen socket for %s\n",
	    thistrans->TransName, 0, 0);
	return NULL;
    }

    /* Save the index for later use */

    return ciptr;
}

static XtransConnInfo
TRANS(MockTCPReopenCLTSServer) (thistrans, fd, port)

Xtransport *thistrans;
int	   fd;
char	   *port;

{
    XtransConnInfo	ciptr;

    PRMSG (2,
	"MockTCPReopenCLTSServer(%d, %s)\n", fd, port, 0);

    if ((ciptr = TRANS(MockTCPReopen) (fd, port)) == NULL)
    {
	PRMSG (1,
	     "MockTCPReopenCLTSServer: Unable to reopen socket for %s\n",
	     thistrans->TransName, 0, 0);
	return NULL;
    }

    /* Save the index for later use */

    return ciptr;
}

#endif /* TRANS_REOPEN */


static int
TRANS(MockTCPSetOption) (ciptr, option, arg)

XtransConnInfo 	ciptr;
int 		option;
int 		arg;

{
    PRMSG (2,"MockTCPSetOption(%d,%d,%d)\n", ciptr->fd, option, arg);

    return -1;
}


#ifdef TRANS_SERVER

static int
TRANS(MockTCPCreateListener) (ciptr, port)

XtransConnInfo	ciptr;
int		port;

{
    int	fd = ciptr->fd;
    int	retry;

    PRMSG (3, "MockTCPCreateListener(%x,%d)\n", ciptr, fd, 0);

    retry = 20;

    while (mt_bind (fd, port) < 0)
    {
	if (errno == EADDRINUSE)
	    return TRANS_ADDR_IN_USE;

	if (retry-- == 0) {
	    PRMSG (1, "MockTCPCreateListener: failed to bind listener\n",
		0, 0, 0);
	    close (fd);
	    return TRANS_CREATE_LISTENER_FAILED;
	}
#ifdef SO_REUSEADDR
	sleep (1);
#else
	sleep (10);
#endif /* SO_REUSEDADDR */
    }

    /*if (mt_listen (fd) < 0)
    {
	PRMSG (1, "MockTCPCreateListener: listen() failed\n", 0, 0, 0);
	close (fd);
	return TRANS_CREATE_LISTENER_FAILED;
    }*/
	
    /* Set a flag to indicate that this connection is a listener */

    ciptr->flags = 1;

    return 0;
}


static int
TRANS(MockTCP_CreateListener) (ciptr, port)

XtransConnInfo 	ciptr;
char 		*port;

{
    struct sockaddr_in	sockname;
    int		namelen = sizeof(sockname);
    int		status;
    short	tmpport;
    _Xgetservbynameparams sparams;
    struct servent *servp;
    int		realport;

#define PORTBUFSIZE	64	/* what is a real size for this? */

    char	portbuf[PORTBUFSIZE];
    
    PRMSG (2, "MockTCPCreateListener(%s)\n", port, 0, 0);

    /*
     * X has a well known port, that is transport dependent. It is easier
     * to handle it here, than try and come up with a transport independent
     * representation that can be passed in and resolved the usual way.
     *
     * The port that is passed here is really a string containing the idisplay
     * from ConnectDisplay().
     */

    if (is_numeric (port))
    {
	tmpport = (short) atoi (port);

	realport=X_MOCKTCP_PORT+tmpport;
    }


    if ((status = TRANS(MockTCPCreateListener) (ciptr, realport)) < 0)
    {
	PRMSG (1,
    "MockTCP_CreateListener: ...MockTCPCreateListener() failed\n",
	    0, 0, 0);
	return status;
    }

    if (TRANS(MockTCPGetAddr) (ciptr) < 0)
    {
	PRMSG (1,
       "MockTCP_CreateListener: ...MockTCPGetAddr() failed\n",
	    0, 0, 0);
	return TRANS_CREATE_LISTENER_FAILED;
    }

    return 0;
}

static XtransConnInfo
TRANS(MockTCPAccept) (ciptr, status)

XtransConnInfo ciptr;
int	       *status;

{
    XtransConnInfo	newciptr;
    struct sockaddr_in	sockname;
    int			namelen = sizeof(sockname);

    PRMSG (2, "MockTCPAccept(%x,%d)\n", ciptr, ciptr->fd, 0);

    if ((newciptr = (XtransConnInfo) xcalloc (
	1, sizeof(struct _XtransConnInfo))) == NULL)
    {
	PRMSG (1, "MockTCPAccept: malloc failed\n", 0, 0, 0);
	*status = TRANS_ACCEPT_BAD_MALLOC;
	return NULL;
    }

    if ((newciptr->fd = mt_accept (ciptr->fd,
	/*(struct sockaddr *)*/ &sockname)) < 0)
    {
	PRMSG (1, "MockTCPAccept: accept() failed\n", 0, 0, 0);
	xfree (newciptr);
	*status = TRANS_ACCEPT_FAILED;
	return NULL;
    }

    /*
     * Get this address again because the transport may give a more 
     * specific address now that a connection is established.
     */

    if (TRANS(MockTCPGetAddr) (newciptr) < 0)
    {
	PRMSG (1,
	    "MockTCPAccept: ...MockTCPGetAddr() failed:\n",
	    0, 0, 0);
	close (newciptr->fd);
	xfree (newciptr);
	*status = TRANS_ACCEPT_MISC_ERROR;
        return NULL;
    }

    if (TRANS(MockTCPGetPeerAddr) (newciptr) < 0)
    {
	PRMSG (1,
	  "MockTCPAccept: ...MockTCPGetPeerAddr() failed:\n",
		0, 0, 0);
	close (newciptr->fd);
	if (newciptr->addr) xfree (newciptr->addr);
	xfree (newciptr);
	*status = TRANS_ACCEPT_MISC_ERROR;
        return NULL;
    }

    *status = 0;

    return newciptr;
}

#endif /* TRANS_SERVER */


#ifdef TRANS_CLIENT

static int
TRANS(MockTCPConnect) (ciptr, host, port)

XtransConnInfo 	ciptr;
char 		*host;
char 		*port;

{
    struct sockaddr_in	sockname;
#if defined(SVR4) || defined(SCO325)
    size_t namelen = sizeof sockname;
#else
    int namelen = sizeof sockname;
#endif
    _Xgethostbynameparams hparams;
    _Xgetservbynameparams sparams;
    struct hostent	*hostp;
    struct servent	*servp;

#define PORTBUFSIZE	64	/* what is a real size for this? */
    char	portbuf[PORTBUFSIZE];

    int			ret;
    short		tmpport;
    unsigned long 	tmpaddr;
    char 		hostnamebuf[256];		/* tmp space */
    int 		realport;

    PRMSG (2,"MockTCPConnect(%d,%s,%s)\n", ciptr->fd, host, port);

    if (!host)
    {
	hostnamebuf[0] = '\0';
	(void) TRANS(GetHostname) (hostnamebuf, sizeof hostnamebuf);
	host = hostnamebuf;
    }

    /*
     * X has a well known port, that is transport dependent. It is easier
     * to handle it here, than try and come up with a transport independent
     * representation that can be passed in and resolved the usual way.
     *
     * The port that is passed here is really a string containing the idisplay
     * from ConnectDisplay().
     */

    if (is_numeric (port))
    {
	tmpport = (short) atoi (port);
	realport=X_MOCKTCP_PORT + tmpport;
    }
    else
    {
	    PRMSG (1,"MockTCPConnect: Invalid port number: %s\n",
		  portbuf, 0, 0);
	    return TRANS_CONNECT_FAILED;
    }

    /*
     * Do the connect()
     */

    if (mt_connect (ciptr->fd, host, realport)<0)
	    return TRANS_CONNECT_FAILED;	
    

    /*
     * Sync up the address fields of ciptr.
     */
    
    if (TRANS(MockTCPGetAddr) (ciptr) < 0)
    {
	PRMSG (1,
	   "MockTCPConnect: ...MockTCPGetAddr() failed:\n",
	   0, 0, 0);
	return TRANS_CONNECT_FAILED;
    }

    if (TRANS(MockTCPGetPeerAddr) (ciptr) < 0)
    {
	PRMSG (1,
	 "MockTCPConnect: ...MockTCPGetPeerAddr() failed:\n",
	      0, 0, 0);
	return TRANS_CONNECT_FAILED;
    }

    return 0;
}

#endif /* TRANS_CLIENT */


static int
TRANS(MockTCPBytesReadable) (ciptr, pend)

XtransConnInfo ciptr;
BytesReadable_t *pend;
{
    PRMSG (2,"MockTCPBytesReadable(%x,%d,%x)\n",
	ciptr, ciptr->fd, pend);
	return mt_bytes_readable(ciptr->fd, pend);
}


static int
TRANS(MockTCPRead) (ciptr, buf, size)

XtransConnInfo	ciptr;
char		*buf;
int		size;

{
    int i;
    PRMSG (2,"MockTCPRead(%d,%x,%d)\n", ciptr->fd, buf, size);

    i=mt_read (ciptr->fd, buf, size);
    PRMSG(2, "MockTCPRead about to return %d\n", i, 0, 0);
    if(i<0)
	perror("mt_read() failed");
    return i;
}


static int
TRANS(MockTCPWrite) (ciptr, buf, size)

XtransConnInfo ciptr;
char 	       *buf;
int 	       size;

{
    PRMSG (2,"MockTCPWrite(%d,%x,%d)\n", ciptr->fd, buf, size);

    return mt_write (ciptr->fd, buf, size);
}


static int
TRANS(MockTCPReadv) (ciptr, buf, size)

XtransConnInfo	ciptr;
struct iovec 	*buf;
int 		size;

{
	int i, bufflen, msglen, bytescopied, len;
	void * temp;
    PRMSG (2,"MockTCPReadv(%d,%x,%d)\n", ciptr->fd, buf, size);

	for(i=0, bufflen=0; i<size; i++) 
		bufflen+=buf[i].iov_len;

	if((temp=malloc(bufflen))==NULL)
	{
		PRMSG(2, "MockTCPReadv: malloc() failed.", 0, 0, 0);
		return -1;		
	}

	msglen=mt_read(ciptr->fd, temp, bufflen);

	for(i=0, bytescopied=0;(bytescopied<msglen) && (i<size); i++)
	{
		len=(msglen-bytescopied<buf[i].iov_len)
			?msglen-bytescopied
			:buf[i].iov_len;
		bcopy(temp+bytescopied, buf[i].iov_base, len);
		bytescopied+=len;
	}

	free(temp);

	return msglen;
}


static int
TRANS(MockTCPWritev) (ciptr, buf, size)

XtransConnInfo 	ciptr;
struct iovec 	*buf;
int 		size;

{
	int i, bufflen, msglen, bytescopied, len;
	void *temp;
    
	PRMSG (2,"MockTCPWritev(%d,%x,%d)\n", ciptr->fd, buf, size);

	for(i=0, bufflen=0; i<size; i++) 
		bufflen+=buf[i].iov_len;

	if((temp=malloc(bufflen))==NULL)
	{
		PRMSG(1, "MockTCPWritev: malloc() failed.", 0, 0, 0);
		return -1;		
	}

	for(i=0, bytescopied=0;i<size; i++)
	{
		bcopy(buf[i].iov_base, temp+bytescopied, buf[i].iov_len);
		bytescopied+=buf[i].iov_len;
	}

	msglen=mt_write(ciptr->fd, temp, bufflen);

	free(temp);
	
	return msglen;
}


static int
TRANS(MockTCPDisconnect) (ciptr)

XtransConnInfo ciptr;

{
    PRMSG (2,"MockTCPDisconnect(%x,%d)\n", ciptr, ciptr->fd, 0);
	return 0;
}


static int
TRANS(MockTCP_Close) (ciptr)

XtransConnInfo ciptr;

{
	int i;
    PRMSG (2,"MockTCP_Close(%x,%d)\n", ciptr, ciptr->fd, 0);
	i=mt_close (ciptr->fd);
    PRMSG (2,"MockTCP_Close about to return %d", i, 0, 0);
    return i;
}



Xtransport	TRANS(MockTCPFuncs) = {
	/* MockTCP Interface */
	"mocktcp",
        0,
#ifdef TRANS_CLIENT
	TRANS(MockTCPOpenCOTSClient),
#endif /* TRANS_CLIENT */
#ifdef TRANS_SERVER
	TRANS(MockTCPOpenCOTSServer),
#endif /* TRANS_SERVER */
#ifdef TRANS_CLIENT
	TRANS(MockTCPOpenCLTSClient),
#endif /* TRANS_CLIENT */
#ifdef TRANS_SERVER
	TRANS(MockTCPOpenCLTSServer),
#endif /* TRANS_SERVER */
#ifdef TRANS_REOPEN
	TRANS(MockTCPReopenCOTSServer),
	TRANS(MockTCPReopenCLTSServer),
#endif
	TRANS(MockTCPSetOption),
#ifdef TRANS_SERVER
	TRANS(MockTCP_CreateListener),
	NULL,		       			/* ResetListener */
	TRANS(MockTCPAccept),
#endif /* TRANS_SERVER */
#ifdef TRANS_CLIENT
	TRANS(MockTCPConnect),
#endif /* TRANS_CLIENT */
	TRANS(MockTCPBytesReadable),
	TRANS(MockTCPRead),
	TRANS(MockTCPWrite),
	TRANS(MockTCPReadv),
	TRANS(MockTCPWritev),
	TRANS(MockTCPDisconnect),
	TRANS(MockTCP_Close),
	TRANS(MockTCP_Close),
	};


